# Copyright 2017 The Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Alias MiniFlow APIs for TensorFlow APIs.

import miniflow as tf
tf.int32
tf.float32
tf.float64
tf.Graph
tf.Session
tf.Variable
tf.placeholder
tf.constant
tf.add
tf.minus
tf.multiple
tf.divide
tf.square
tf.global_variables_initializer
tf.local_variables_initialize
tf.get_variable
"""

# TODO: Need to import all after installation
from . import graph
from . import session
from . import ops
from . import optimizer
from . import train
from . import webhdfs

int32 = int
float32 = float
float64 = float

Graph = graph.Graph

Session = session.Session

Variable = ops.VariableOp
placeholder = ops.PlaceholderOp
constant = ops.ConstantOp
add = ops.AddOp
minus = ops.MinusOp
multiple = ops.MultipleOp
divide = ops.DivideOp
square = ops.SquareOp
global_variables_initializer = ops.GlobalVariablesInitializerOp
local_variables_initializer = ops.LocalVariablesInitializerOp
get_variable = ops.get_variable


# Copyright 2017 The Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Access HDFS weith WebHDFs. Need to enable webhdfs in hdfs-site.xml.

<property>
  <name>dfs.webhdfs.enabled</name>
  <value>true</value>
</property> 
"""

import os
import logging
# TODO: Not compatible for python 3
# import urlparse
# import httplib

logger = logging.getLogger(name="webhdfs")


class Webhdfs(object):
  """
  The client to access WebHDFS. More usage refer to
  https://github.com/drelu/webhdfs-py/blob/master/webhdfs/webhdfs.py.
  """

  def __init__(self, namenode_host, namenode_port, username, timeout=60):
    self._namenode_host = namenode_host
    self._namenode_port = namenode_port
    self._username = username
    self._timeout = timeout
    self._webhdfs_prefix = "/webhdfs/v1"

  """
  def get(self, source_path):
    if os.path.isabs(source_path) == False:
      raise Exception("Only absolute paths supported: %s" % source_path)

    url_path = self._webhdfs_prefix + source_path + '?op=OPEN&overwrite=true&user.name=' + self._username
    logger.debug("GET URL: %s" % url_path)

    httpClient = httplib.HTTPConnection(
        self._namenode_host, self._namenode_port, timeout=self._timeout)
    httpClient.request('GET', url_path, headers={})
    response = httpClient.getresponse()
    data = None

    if response.length != None:
      msg = response.msg
      redirect_location = msg["location"]
      logger.debug("HTTP Response: %d, %s" % (response.status,
                                              response.reason))
      logger.debug("HTTP Location: %s" % (redirect_location))
      result = urlparse.urlparse(redirect_location)
      redirect_host = result.netloc[:result.netloc.index(":")]
      redirect_port = result.netloc[(result.netloc.index(":") + 1):]
      redirect_path = result.path + "?" + result.query
      logger.debug("Send redirect to: host: %s, port: %s, path: %s " %
                   (redirect_host, redirect_port, redirect_path))
      fileDownloadClient = httplib.HTTPConnection(
          redirect_host, redirect_port, timeout=600)
      fileDownloadClient.request('GET', redirect_path, headers={})
      response = fileDownloadClient.getresponse()
      logger.debug("HTTP Response: %d, %s" % (response.status,
                                              response.reason))
      data = response.read()

    httpClient.close()
    return data
    """


def test_Webhdfs():
  webhdfs = Webhdfs("localhost", 50070, "tobe")



# Copyright 2017 The Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from session import Session
import ops


class SessionTest(unittest.TestCase):
  def setUp(self):
    pass

  def tearDown(self):
    pass

  def test_init(self):
    foo = Session()
    self.assertEqual(foo.__class__, Session)

  def test_run(self):
    pass


if __name__ == '__main__':
  unittest.main()



# Copyright 2017 The Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import graph
from . import session
from . import ops


class OptimizerMinimizeOp(ops.Op):
  def __init__(self, optimizer, loss, name="OptimizerMinimize"):
    super(OptimizerMinimizeOp, self).__init__()

    self._optimizer = optimizer
    self._loss = loss

    self._graph = loss._graph
    self._graph.add_to_graph(self)

  def forward(self):
    variablename_grad_map = self._optimizer.compute_gradients(self._loss)
    self._optimizer.apply_gradients(variablename_grad_map)

  def grad(self):
    raise NotImplementedError


class Optimizer(object):
  def __init__(self, name="Optimizer"):
    self.name = name

  def minimize(self, loss):
    pass

  def compute_gradients(self):
    pass

  def apply_gradients(self):
    pass

  def get_name(self):
    return self.name

  def set_name(self, name):
    self.name = name


class GradientDescentOptimizer(Optimizer):
  def __init__(self, learning_rate=0.01, name="GradientDescent"):
    super(GradientDescentOptimizer, self).__init__(name)

    self._learning_rate = learning_rate

    self._graph = graph.get_default_graph()

  def get_learning_rate(self):
    return self._learning_rate

  def set_learning_rate(self, learning_rate):
    # TODO: Check type of parameter
    self._learning_rate = learning_rate

  def get_graph(self):
    return self._graph

  def set_graph(self, graph):
    self._graph = graph

  def minimize(self, loss, global_step=None):
    return OptimizerMinimizeOp(self, loss)

  def compute_gradients(self, loss):

    variablename_variable_map = self._graph.get_trainable_variables_collection(
    )

    variablename_grad_map = {}
    for variable_name, variable in variablename_variable_map.items():
      grad = loss.grad(variable_name)
      variablename_grad_map[variable_name] = grad

    return variablename_grad_map

  def apply_gradients(self, variablename_grad_map):

    variablename_variable_map = self._graph.get_trainable_variables_collection(
    )

    for variable_name, variable in variablename_variable_map.items():
      grad = variablename_grad_map[variable_name]
      final_grad = self._learning_rate * grad
      variable.set_value(variable.get_value() - final_grad)


class AdagradOptimizer(Optimizer):
  pass


class MomentumOptimizer(Optimizer):
  pass


class AdamOptimizer(Optimizer):
  def __init__(self,
               learning_rate=0.01,
               beta1=0.9,
               beta2=0.99,
               epsilon=0.01,
               name="Adam"):
    super(GradientDescentOptimizer, self).__init__(name)
    self._learning_rate = learning_rate
    self._beta1 = beta1
    self._beta2 = beta2
    self._epsilon = epsilon
    self._m = []
    self._v = []



# Copyright 2017 The Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import graph
import ops
from graph import Graph


class GlobalTest(unittest.TestCase):
  def setUp(self):
    pass

  def tearDown(self):
    pass

  def test_get_default_graph(self):
    graph1 = graph.get_default_graph()
    graph2 = graph.get_default_graph()

    self.assertEqual(graph1.__class__, Graph)
    self.assertEqual(graph2.__class__, Graph)
    self.assertEqual(graph1, graph2)


class GraphTest(unittest.TestCase):
  def setUp(self):
    pass

  def tearDown(self):
    pass

  def test_init(self):
    graph = Graph()
    self.assertEqual(graph.__class__, Graph)

  def test_get_name_op_map(self):
    graph = Graph()
    op = ops.Op()
    graph.add_to_graph(op)
    name_op_map = graph.get_name_op_map()
    self.assertEqual(name_op_map.keys()[0], op.get_name())
    self.assertEqual(name_op_map.values()[0], op)


if __name__ == '__main__':
  unittest.main()



# Copyright 2017 The Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from ops import Op
from ops import VariableOp
from ops import PlaceholderOp
from ops import SquareOp


class OpTest(unittest.TestCase):
  def setUp(self):
    pass

  def tearDown(self):
    pass

  def test_init(self):
    foo = Op()
    self.assertEqual(foo.__class__, Op)

  def test_forward(self):
    foo = Op()

    with self.assertRaises(NotImplementedError):
      foo.forward()

  def test_grad(self):
    foo = Op()

    with self.assertRaises(NotImplementedError):
      foo.grad()


class SquareOpTest(unittest.TestCase):
  def test_SquareOp(self):
    w = VariableOp(10)
    b = VariableOp(20)
    x = PlaceholderOp(float)
    x.set_value(2.0)
    y = PlaceholderOp(float)
    y.set_value(3.0)
    """
    loss = SquareOp(y - (w * x + b))
    print("w: {}, forward: {}, grad: {}".format(w.get_value(),
                                                loss.forward(),
                                                loss.grad(w.get_name())))  # 148.0
    print("b: {}, forward: {}, grad: {}".format(b.get_value(),
                                                loss.forward(),
                                                loss.grad(b.get_name())))  # 74.0
    """


if __name__ == '__main__':
  unittest.main()



# Copyright 2017 The Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import ops


class Session(object):
  """The session to run specified op from specified graph."""

  def __init__(self):
    pass

  def __enter__(self):
    """Support with statement."""
    return self

  def __exit__(self, type, value, trace):
    pass

  def run(self, op, feed_dict=None, options=None):

    # Update the value of PlaceholerOp with feed_dict data
    name_op_map = op._graph.get_name_op_map()

    if feed_dict != None:
      # Example: {"Placeholer_1": 10} or {PlaceholderOp: 10}
      for op_or_opname, value in feed_dict.items():
        if isinstance(op_or_opname, str):
          placeholder_op = name_op_map[op_or_opname]
        else:
          placeholder_op = op_or_opname

        if isinstance(placeholder_op, ops.PlaceholderOp):
          placeholder_op.set_value(value)

    result = op.forward()
    return result



# Copyright 2017 The Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from optimizer import Optimizer
from optimizer import GradientDescentOptimizer


class OptimizerTest(unittest.TestCase):
  def setUp(self):
    pass

  def tearDown(self):
    pass

  def test_init(self):
    foo = Optimizer()
    self.assertEqual(foo.__class__, Optimizer)

  def test_get_name(self):
    foo = Optimizer("FooOptimizer")
    self.assertEqual(foo.get_name(), "FooOptimizer")

  def test_get_name(self):
    foo = Optimizer()
    foo.set_name("FooOptimizer")
    self.assertEqual(foo.get_name(), "FooOptimizer")


class GradientDescentOptimizerTest(unittest.TestCase):
  def setUp(self):
    pass

  def tearDown(self):
    pass

  def test_init(self):
    foo = GradientDescentOptimizer()
    self.assertEqual(foo.__class__, GradientDescentOptimizer)

  def test_compute_gradients(self):
    foo = GradientDescentOptimizer(learning_rate=0.1)
    """ TODO: Add more tests
    optimizer.minimize(loss)
    sess = session.Session()
    sess.run(optimizer)
    """


if __name__ == '__main__':
  unittest.main()



# Copyright 2017 The Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains all the basic operations."""

from abc import ABCMeta, abstractmethod
import logging
import math
import os
import sys

#from . import graph
import graph

# Enable swig by environment variable
if "ENABLE_SWIG_OP" in os.environ:
  logging.info("Enable swig operations by environment variable")
  sys.path.append("../")
  import swig.op


class Op(object):
  """The basic class for all operation."""

  def __init__(self, name="Op"):
    # Be compatiable with TensorFlow to remove underline
    self.name = name

  def get_name(self):
    return self.name

  def set_name(self, name):
    self.name = name

  @abstractmethod
  def forward(self):
    # TODO: No need to implement in abstract method
    raise NotImplementedError

  def grad(self):
    raise NotImplementedError

  def __add__(self, other):
    return AddOp(self, other)

  def __radd__(self, other):
    return self.__add__(other)

  def __sub__(self, other):
    return MinusOp(self, other)

  def __rsub__(self, other):
    return MinusOp(other, self)

  def __mul__(self, other):
    return MultipleOp(self, other)

  def __rmul__(self, other):
    return self.__mul__(other)

  def __div__(self, other):
    return DivideOp(self, other)

  def __rdiv__(self, other):
    return DivideOp(other, self)

  def __pow__(self, power, modulo=None):
    return PowerOp(self, power)


class PlaceholderOp(Op):
  """The placeholer operation which value is set when Session.run()"""

  def __init__(self, dtype=None, shape=None, name="Placeholder"):
    super(PlaceholderOp, self).__init__(name)
    # TODO: Use dtype and shape
    self._dtype = dtype
    self._shape = shape

    # The value is None util Session.run() with feed_dict parameter
    self._value = None

    # TODO: Support other graph instance
    self._graph = graph.get_default_graph()
    self._graph.add_to_graph(self)

  def set_value(self, value):
    self._value = value

  def get_value(self):
    return self._value

  def forward(self):
    return self._value

  def grad(self, partial_derivative_opname=None):
    return 0


class ConstantOp(Op):
  """The constant operation which contains one initialized value."""

  def __init__(self, value, name="Constant"):
    super(ConstantOp, self).__init__(name)
    self._value = value

    self._graph = graph.get_default_graph()
    self._graph.add_to_graph(self)

  # TODO: Not allow to set the value

  def get_value(self):
    return self._value

  def forward(self):
    return self._value

  def grad(self, partial_derivative_opname=None):
    return 0


class VariableOp(Op):
  """
  The variable operation which contains one variable. The variable may be
  trainable or not-trainable. This is used to define the machine learning
  models.
  """

  def __init__(self, value, is_trainable=True, name="Variable"):
    super(VariableOp, self).__init__(name)
    self._value = value
    self._is_trainable = is_trainable

    self._graph = graph.get_default_graph()
    self._graph.add_to_graph(self)

    if self._is_trainable:
      self._graph.add_to_trainable_variables_collection(self.get_name(), self)

  def get_value(self):
    return self._value

  def set_value(self, value):
    self._value = value

  def forward(self):
    return self._value

  def grad(self, partial_derivative_opname=None):
    if partial_derivative_opname is None:
      grad = 1
    else:
      if self.name == partial_derivative_opname:
        # Specify to compute this derivative
        grad = 1
      else:
        # Specify to compute other derivative
        grad = 0
    return grad


class PowerOp(Op):
  def __init__(self, input, power, name="Power"):
    super(PowerOp, self).__init__(name)

    if not isinstance(input, Op):
      self._op = ConstantOp(input)
    else:
      self._op = input

    self._power = power

    self._graph = graph.get_default_graph()
    self._graph.add_to_graph(self)

  def forward(self):
    result = pow(self._op.forward(), self._power)
    return result

  def grad(self, partial_derivative_opname=None):
    if isinstance(self._op, PlaceholderOp) or isinstance(self._op, ConstantOp):
      # op is the constant
      grad = 0
    elif isinstance(self._op, VariableOp):
      # op is the variable
      grad = self._power * pow(self._op.forward(), self._power - 1)
    else:
      # op is other complex operation and use chain rule
      grad = self._power * pow(self._op.forward(), self._power - 1
                               ) * self._op.grad(partial_derivative_opname)

    return grad


class SquareOp(PowerOp):
  def __init__(self, input, name="Square"):
    super(SquareOp, self).__init__(input, 2, name)


class SquareOpOld(Op):
  # TODO: Deprecated op

  def __init__(self, input, name="Square"):
    if not isinstance(input, Op):
      self.op = ConstantOp(input)
    else:
      self.op = input

    self.name = name

    self.graph = graph.get_default_graph()
    self.graph.add_to_graph(self)

  def forward(self):
    if "ENABLE_SWIG_OP" in os.environ:
      result = swig.op.square(self.op.forward())
    else:
      result = pow(self.op.forward(), 2)
    return result

  def grad(self, partial_derivative_opname=None):
    if isinstance(self.op, PlaceholderOp) or isinstance(self.op, ConstantOp):
      # op is the constant
      grad = 0
    elif isinstance(self.op, VariableOp):
      # op is the variable
      if "ENABLE_SWIG_OP" in os.environ:
        grad = swig.op.multiple(2, self.op.forward())
      else:
        grad = 2 * self.op.forward()
    else:
      # op is other complex operation and use chain rule
      grad = 2 * self.op.forward() * self.op.grad(partial_derivative_opname)

    return grad


class CubicOp(PowerOp):
  def __init__(self, input, name="Cubic"):
    super(CubicOp, self).__init__(input, 3, name)


class CubicOpOld(Op):
  # TODO: Deprecated op

  def __init__(self, input, name="Cubic"):
    if not isinstance(input, Op):
      self.op = ConstantOp(input)
    else:
      self.op = input

    self.name = name

    self.graph = graph.get_default_graph()
    self.graph.add_to_graph(self)

  def forward(self):
    if "ENABLE_SWIG_OP" in os.environ:
      result = swig.op.cubic(self.op.forward())
    else:
      result = math.pow(self.op.forward(), 3)
    return result

  def grad(self, partial_derivative_opname=None):
    if isinstance(self.op, PlaceholderOp) or isinstance(self.op, ConstantOp):
      # op is the constant
      grad = 0
    elif isinstance(self.op, VariableOp):
      # op is the variable
      if "ENABLE_SWIG_OP" in os.environ:
        grad = swig.op.multiple(3, swig.op.square(self.op.forward()))
      else:
        grad = 3 * math.pow(self.op.forward(), 2)
    else:
      # op is other complex operation
      grad = 3 * math.pow(self.op.forward(),
                          2) * self.op.grad(partial_derivative_opname)

    return grad


def SigmoidOp(Op):
  # TODO: Need to implement the forward and grad functions

  def __init__(self, input, name="Sigmoid"):
    super(SigmoidOp, self).__init__(name)


class AddOp(Op):
  """
  The addition operation which has only two inputs. The input can be
  primitive, ConstantOp, PlaceholerOp, VariableOp or other ops.
  """

  def __init__(self, input1, input2, name="Add"):
    super(AddOp, self).__init__(name)

    if not isinstance(input1, Op):
      self._op1 = ConstantOp(input1)
    else:
      self._op1 = input1

    if not isinstance(input2, Op):
      self._op2 = ConstantOp(input2)
    else:
      self._op2 = input2

    self._graph = graph.get_default_graph()
    self._graph.add_to_graph(self)

  def forward(self):
    result = self._op1.forward() + self._op2.forward()
    return result

  def grad(self, partial_derivative_opname=None):
    result = self._op1.grad(partial_derivative_opname) + self._op2.grad(
        partial_derivative_opname)
    return result


class MinusOp(Op):
  """
  The minus operation.
  """

  def __init__(self, input1, input2, name="Minus"):
    super(MinusOp, self).__init__(name)

    if not isinstance(input1, Op):
      self._op1 = ConstantOp(input1)
    else:
      self._op1 = input1

    if not isinstance(input2, Op):
      self._op2 = ConstantOp(input2)
    else:
      self._op2 = input2

    self._graph = graph.get_default_graph()
    self._graph.add_to_graph(self)

  def forward(self):
    result = self._op1.forward() - self._op2.forward()
    return result

  def grad(self, partial_derivative_opname=None):
    result = self._op1.grad(partial_derivative_opname) - self._op2.grad(
        partial_derivative_opname)
    return result


class AddNOp(Op):
  def __init__(self, *inputs):
    # TODO: Deprecated op
    # TODO: Support user defined name in the parameter
    self.name = "AddN"

    self.ops = []
    for input in inputs:
      if not isinstance(input, Op):
        input = ConstantOp(input)
      self.ops.append(input)

    self.graph = graph.get_default_graph()
    self.graph.add_to_graph(self)

  def forward(self):
    result = 0
    for op in self.ops:
      result += op.forward()
    return result

  def grad(self, partial_derivative_opname=None):
    result = 0
    for op in self.ops:
      result += op.grad(partial_derivative_opname)
    return result


# TODO: Can not support operations like "x * x", only "x * 3"
class MultipleOp(Op):
  def __init__(self, input1, input2, name="Multiple"):
    super(MultipleOp, self).__init__(name)

    if not isinstance(input1, Op):
      self._op1 = ConstantOp(input1)
    else:
      self._op1 = input1

    if not isinstance(input2, Op):
      self._op2 = ConstantOp(input2)
    else:
      self._op2 = input2

    self._graph = graph.get_default_graph()
    self._graph.add_to_graph(self)

  def forward(self):
    result = self._op1.forward() * self._op2.forward()
    return result

  def grad(self, partial_derivative_opname=None):

    if isinstance(self._op1, PlaceholderOp) or isinstance(
        self._op1, ConstantOp):
      # op1 is the coefficient of this formula
      op1_grad = self._op1.forward()

      if isinstance(self._op2, PlaceholderOp) or isinstance(
          self._op2, ConstantOp):
        # two elements are both constant values
        op2_grad = 0
      else:
        # op2 may has VariableOp
        op2_grad = self._op2.grad(partial_derivative_opname)

    elif isinstance(self._op2, PlaceholderOp) or isinstance(
        self._op2, ConstantOp):
      # op2 is the coefficient of this formula
      op2_grad = self._op2.forward()

      # op1 may has VariableOp
      op1_grad = self._op1.grad(partial_derivative_opname)

    else:
      # op1 and op2 may has VariableOp
      logging.error(
          "Not support complex formula which has multiple VariableOp")
      raise NotImplementedError

    result = op1_grad * op2_grad
    return result


class MultipleNOp(Op):
  # TODO: Deprecated op
  """The multiple operation for n inputs."""

  def __init__(self, *inputs):
    self.name = "MultipleN"

    self.ops = []
    for input in inputs:
      if not isinstance(input, Op):
        input = ConstantOp(input)
      self.ops.append(input)

    self.graph = graph.get_default_graph()
    self.graph.add_to_graph(self)

  def forward(self):
    result = 1
    for op in self.ops:
      result *= op.forward()
    return result

  def grad(self, partial_derivative_opname=None):
    # TODO: Check the type of op to compute gradients
    result = 1
    for op in self.ops:
      result *= op.grad(partial_derivative_opname)
    return result


class DivideOp(Op):
  def __init__(self, input1, input2, name="Divide"):
    super(DivideOp, self).__init__(name)

    if not isinstance(input1, Op):
      self._op1 = ConstantOp(input1)
    else:
      self._op1 = input1

    if not isinstance(input2, Op):
      self._op2 = ConstantOp(input2)
    else:
      self._op2 = input2

    self._graph = graph.get_default_graph()
    self._graph.add_to_graph(self)

  def forward(self):
    result = self._op1.forward() / self._op2.forward()
    return result

  def grad(self, partial_derivative_opname=None):
    result = self._op1.grad(partial_derivative_opname) / self._op2.grad(
        partial_derivative_opname)
    return result


class UpdateVariableOp(Op):
  # TODO: Deprecated op

  def __init__(self, variableOp, value, name="UpdateVariableOp"):
    self.variableOp = variableOp
    self.value = value
    self.name = name

    self.graph = graph.get_default_graph()
    self.graph.add_to_graph(self)

  def forward(self):
    self.variableOp.set_value(self.value)
    return self.value

  # TODO: Add grad() if needed


class UpdateVariableNOp(Op):
  # TODO: Deprecated op

  def __init__(self, variableop_value_map, name="UpdateVariableNOp"):
    self.variableop_value_map = variableop_value_map
    self.name = name

    self.graph = graph.get_default_graph()
    self.graph.add_to_graph(self)

  def forward(self):
    # TODO: Need to test the loop
    for variableOp, value in enumerate(self.variableop_value_map):
      variableOp.set_value(value)
    return self.variableop_value_map


class GlobalVariablesInitializerOp(Op):
  def __init__(self, name="GlobalVariablesInitializer"):
    super(GlobalVariablesInitializerOp, self).__init__(name)

    self._graph = graph.get_default_graph()
    self._graph.add_to_graph(self)

  def forward(self):
    pass

  def grad(self):
    raise NotImplementedError


class LocalVariablesInitializerOp(Op):
  def __init__(self, name="LocalVariablesInitializer"):
    super(LocalVariablesInitializerOp, self).__init__(name)

    self._graph = graph.get_default_graph()
    self._graph.add_to_graph(self)

  def forward(self):
    pass

  def grad(self):
    raise NotImplementedError


def get_variable(name="Variable",
                 value=None,
                 shape=None,
                 dtype=None,
                 initializer=None,
                 regularizer=None,
                 reuse=None,
                 trainable=True):
  # TODO: Support default graph only
  _graph = graph.get_default_graph()

  if name in _graph.get_name_op_map():
    return _graph.get_name_op_map()[name]
  else:
    variable = VariableOp(value=value, name=name)
    return variable
